/*
 *	This software is Copyright (c) 1990 by Ed Braaten.
 *
 *	Permission is hereby granted to copy, reproduce, redistribute or
 *	otherwise use this software as long as: there is no monetary
 *	profit gained specifically from the use or reproduction or this
 *	software, it is not sold, rented, traded or otherwise marketed, and
 *	this copyright notice is included prominently in any copy
 *	made.
 *
 *	The author make no claims as to the fitness or correctness of
 *	this software for any use whatsoever, and it is provided as is. 
 *	Any use of this software is at the user's own risk.
 *
 */

/*	@(#) uubatch.c	1.0	12/31/90 	*/
 
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include "config.h"


/* Global flags & vars */
static int debugflg = 0; 
static int verboseflg = 0;
static int bit12flg = 0;
static int bitgzip = 0;
static int bithgzip = 0;
static int curhour;
static int servcnt;
static int compressio[2];
static int uuxio[2];
static char *servlist = (char *) 0;


void print_usage()
{
   printf("Usage:  uubatch  [-s service(s) ] [ -d ]\n");
}

/* returns the current hour as an integer */
int gethour()
{
   char *ctime();
   long curtime;

   time(&curtime); 
   return(atoi(ctime(&curtime)+11));
}



/* returns day of month as an integer */
int getday()
{
   char *ctime();
   long curtime;

   time(&curtime); 
   return(atoi(ctime(&curtime)+8));
}



/* Returns true (1) if the current weekday matches.
 * Sun = '0', Mon = '1', etc.
*/
int matchwkday(wkday)
char wkday;
{
   static char wkdays[] = "SunMonTueWedThuFriSat";
   char *ctime(), *ptr;
   long curtime;

   /* 'convert' wkday to text */ 
   ptr = wkdays + ((wkday - '0') * 3);
  
   /* compare with the actual weekday */ 
   time(&curtime); 
   if (strncmp(ctime(&curtime), ptr, 3) == 0)
      return(1);
   else 
      return(0);
}



/* Check days field of a site entry.  Returns the count of
 * true conditions.
*/
int chkdays(days)
char *days;
{
   int matchcnt;

   /* go through the days */
   for (matchcnt=0; *days; days++) {
      switch (*days) {
         case 'e':
            if (!(getday() % 2)) {
               if (debugflg)
                  printf("      Day field entry 'e' matches.\n");
               matchcnt++;
            }
            else
               if (debugflg)
                  printf("      Day field entry 'e' doesn't match.\n");
            break;
         case 'o':
            if ((getday() % 2)) {
               if (debugflg)
                  printf("      Day field entry 'o' matches.\n");
               matchcnt++;
            }
            else
               if (debugflg)
                  printf("      Day field entry 'o' doesn't match.\n");
            break;
         case '1':
         case '2':
         case '3':
         case '4':
         case '5':
         case '6':
         case '7':
            if (matchwkday(*days)) {
               if (debugflg)
                  printf("      Day field entry '%c' matches.\n", *days);
               matchcnt++;
            }
            else
               if (debugflg)
                  printf("      Day field entry '%c' doesn't match.\n", *days);
            break;
         case '*':
            matchcnt++; 
            if (debugflg)
               printf("      Day field entry '*' matches.\n");
            break;
         default:
            printf("*** Unknown flag '%c' ignored!\n", *days);   
            break;
      }
   }

   /* all done */
   return(matchcnt);
}

 

/* This routine fills poll_list[] with the site info for sites
 * which have an hour entry matching the current hour.
*/ 
int buildlist()
{
   FILE *fp;
   static char buf[LINEBUFSIZ];
   char *p;
   int linecnt, match, curhour, jobcnt;
   char *sitename, *services, *days, *hours;

   /* open up the POLLFILE */
   if ((fp = fopen(POLLFILE, "r")) == NULL) {
      printf("ERROR!  Can't open %s.  Aborting.\n", POLLFILE);
      exit(-1);
   }

   /* get current hour */
   curhour = gethour();

   if (debugflg)
      printf("Parsing %s for entries with hour '%d'.\n", POLLFILE, curhour); 

   /* parse the lines */
   linecnt = 0, jobcnt = 0;
   while (fgets(buf, sizeof(buf), fp) != NULL) {

      linecnt++;

      /* skip comments & blank lines */
      if ((buf[0] == '\n') || (buf[0] == '#'))
         continue;

      /* set ptrs to fields */
      sitename = strtok(buf, ":"); 
      services = strtok((char *) 0, ":");
      days = strtok((char *) 0, ":");
      hours = strtok((char *) 0, "\n");
      if (debugflg) 
         printf("   At line %d, sitename: %s, services: %s, days: %s\n", 
            linecnt, sitename, services, days);

      /* sanity checks */
      if ((sitename == (char *) 0) || (services == (char *) 0) || 
         (days == (char *) 0) || (hours == (char *) 0)) {
         printf("ERROR parsing line %d in %s.  Aborting.\n",
            linecnt, POLLFILE);
         fclose(fp);
         exit(-1);
      }

      /* chk for an hour entry == curhour if days field valid */
      match = 0;
      if (chkdays(days)) {
         for (p=strtok(hours," \n"); p!=(char*) 0; p=strtok((char*) 0," \n")) {
    
            if (debugflg) 
               printf("      Looking at hour '%s'...", p);
 
            if (atoi(p) == curhour) {
               if (debugflg)
                  printf(" It matches!\n");
               match++;
               break;
            }
            else
               if (debugflg)
                  putchar('\n');
         }
      }

      /* fill in structure if entry found */
      if (match) {

         strcpy(poll_list[jobcnt].sitename, sitename);
         strcpy(poll_list[jobcnt].services, services);

         /* chatter a bit */
         if (debugflg) 
            printf("      *** Match for site %s ***\n", sitename);

         /* note the entry */
         jobcnt++;
      } 
      else {
         if (debugflg)
            printf("      *** No match found for site %s ***\n", sitename);
      }
   }

   /* debug output */
   if (debugflg)
      printf("Parsing completed, %d %s found.\n", 
         jobcnt, (jobcnt==1) ? "job":"jobs"); 

   return(jobcnt);
}



int getservices()
{
   FILE *fp;
   static char buf[LINEBUFSIZ];
   int linecnt, i;

   /* open up the SERVFILE */
   if ((fp = fopen(SERVFILE, "r")) == NULL) {
      printf("ERROR!  Can't open %s.  Aborting.\n", SERVFILE);
      exit(-1);
   }

   if (debugflg)
      printf("Parsing %s for service commands.\n", SERVFILE); 

   /* parse the lines */
   linecnt = 0, i=0;
   while (fgets(buf, sizeof(buf), fp) != NULL) {

      linecnt++;

      /* skip comments & blank lines */
      if ((buf[0] == '\n') || (buf[0] == '#'))
         continue;

      /* sanity check - skip extra lines if they exceed our 
       * configured array size...
      */
      if (i >= NUMSERVICES) {
         printf("Count of services in %s exceeds number of services\n",
            SERVFILE);
         printf("allowed (%d).  Services after and including line %d\n",
            NUMSERVICES, linecnt);
         printf("will be skipped.\n");         
         return(i);
      }

      /* don't allow 'p' as a service id */
      if (buf[0] == 'p') {
         printf("ERROR (line %d):  'p' is not a valid service id.  Aborting.\n",
            linecnt);
         exit(-1);
      }

      /* add entry to list */
      serv_list[i].id = buf[0];
      buf[strlen(buf)-1] = '\0';
      if ((serv_list[i].cmd = malloc(strlen(&buf[2])+1)) == NULL) {
         printf("Can't allocate memory for cmd in line %d of %s.  Aborting.\n",
            linecnt, SERVFILE);
         exit(-1);
      }
      strcpy(serv_list[i].cmd, &buf[2]);

      /* chatter for debugging */
      if (debugflg) {
         printf("   Service id: %c\n", serv_list[i].id);
         printf("          cmd: %s\n", serv_list[i].cmd); 
      }

      /* bump cnt of services */
      i++;

   }

   if (debugflg)
      printf("Parsing completed, %d service %s found.\n", 
         i, (i==1) ? "command":"commands"); 
   
   /* return number of services entered in the list */ 
   return(i); 
}



/* This routine creates a uucp "poll" file in the 
 * spool directory /usr/spool/uucp/"sitename".
*/
makepollfile(sitename)
char *sitename;
{
   static char buf[MAXPATHLEN];
   int omask;
   struct stat sbuf;

   /* clear our umask */
   omask = umask(0);

   /* make sure the spool directory for sitename exists */
   sprintf(buf, "/usr/spool/uucp/%s", sitename);
   if (stat(buf, &sbuf) < 0) {
      mkdir(buf, 0755);
      chown(buf, UUCPUID, UUCPGID);
      if (verboseflg)
         printf("   Created directory '%s'\n", buf);
   }

   /* create the poll file */
   sprintf(buf,"%s/C.%.7sn0000", buf, sitename);
   creat(buf, 0000644);
   chown(buf, UUCPUID, UUCPGID);
   if (verboseflg)
      printf("   Created poll file '%s'\n", buf);

   /* restore our umask */
   umask(omask);
}



/* This routine starts up the "| compress | uux" pipeline...  
 * If something fails, it returns non-zero.
*/
int do_shelloil(sitename, fp)
char *sitename;
FILE *fp;
{
   int pid;
   char buf[SITENAMELEN+8];

   /* open pipes */
   if (pipe(compressio)) {
      printf("Couldn't open pipe for compress!\n");
      return(-1);
   }
   if (pipe(uuxio)) {
      printf("Couldn't open pipe for uux!\n");
      close(compressio[0]), close(compressio[1]);
      return(-1);
   }

   /* write header data into pipe */
   if (bitgzip)
      sprintf(buf, "#! gunbatch\n");
   else
      sprintf(buf, "#! cunbatch\n");
   if (write(uuxio[1], buf, 12) != 12) {
      printf("Error writing to uux pipe!\n");
      return(-1);
   } 

   /* startup compress command */
   switch (fork()) {
      case 0:
         /* set up stdin/stdout */
         close(0), dup(compressio[0]), close(compressio[0]);
         close(1), dup(uuxio[1]), close(uuxio[1]);

         /* don't leave any file descriptors hanging around */
         close(compressio[1]);
         close(uuxio[0]);
         fclose(fp);

         /* become compress... */ 
         if (bit12flg)
	 {
            execle(COMPRESS, "compress", "-b12", (char *) 0, environ);
	    return(0);
	 }
         if (bitgzip || bithgzip )
            execle(COMPRESS2, "gzip", "-S .z ", (char *) 0, environ);
         else
            execle(COMPRESS, "compress",  (char *) 0, environ);
         printf("Exec of %s failed!\n", COMPRESS);
         return(-1);
      case -1:
         printf("Cannot fork for compress!\n");
         return(-1);
      default:
         break;
   }

   /* startup uux command */
   switch (fork()) {
      case 0:
         /* set up stdin - leave stdout as our stdout? */
         close(0), dup(uuxio[0]), close(uuxio[0]);

         /* don't leave any file descriptors hanging around */
         close(compressio[0]);
         close(compressio[1]);
         close(uuxio[1]);
         fclose(fp);
          
         /* make ourselves the news user */
         setuid(NEWSUID);
         setgid(NEWSGID);

         /* become uux... */ 
         sprintf(buf, "%s!rnews", sitename);
#ifdef GOPT
         execle(UUX, "uux",GOPT,"-r","-",buf, (char *) 0, environ);
#else
         execle(UUX, "uux","-r","-",buf, (char *) 0, environ);
#endif
         printf("Exec of %s failed!\n", UUX);
         return(-1);
      case -1:
         printf("Cannot fork for uux!\n");
         return(-1);
      default:
         break;
   }

   /* close unneeded pipehandles */
   close(uuxio[0]), close(uuxio[1]);
   close(compressio[0]);


   /* pipeline is in place */
   return(0);
}



/* This routine batches news for the given site.
 * It is compatible with the Bnews 2.11 batch command.
 * Thanks to Rick Adams for his coding example. ;-)
*/
do_batch(sitename)
char *sitename;
{
   struct stat sbuf;
   extern int errno;
   int i, c, artfd;
   long bytecnt;
   FILE *fp;
   char *p;
   static char buf[BUFSIZ];
   static char batchfile[MAXPATHLEN], wrkfile[MAXPATHLEN];
   static char fname[MAXPATHLEN];

   /* Chk if a batchfile/workfile exists - if workfile
    * already exists, use it...
   */
   sprintf(batchfile, "%s/%s", NEWSBATCHDIR, sitename);
   sprintf(wrkfile, "%s/%s.work", NEWSBATCHDIR, sitename);
   if (access(wrkfile, 0) < 0) {
      if (access(batchfile, 0) < 0 && errno == ENOENT) {
         if (verboseflg) 
            printf("No news to batch for site %s.\n", sitename);
         return;   
      }
      if (rename(batchfile, wrkfile)) {
         printf("Couldn't rename %s to %s!\n", batchfile, wrkfile);
         return; 
      }
   }

   /* open the workfile */
   if ((fp = fopen(wrkfile, "r")) == NULL) {
      printf("Couldn't open %s!  Aborting news batching for %s.\n", 
         wrkfile, sitename);
      return; 
   }

   /* startup the pipeline... */
   if (do_shelloil(sitename, fp)) {
      close(compressio[0]), close(compressio[1]);
      close(uuxio[0]), close(uuxio[1]);
      fclose(fp);
      printf("Aborting news batching for %s.\n", sitename);
      return;
   }

   bytecnt = 0L;
   while (fgets(fname, sizeof(fname), fp) != NULL) {

      /* open the article file */
      fname[strlen(fname)-1] = '\0';
      if ((artfd = open(fname, O_RDONLY)) < 0) {
         if (verboseflg)
            printf("Can't open %s.\n", fname);
         continue;
      }

      /* stat the file and get its size */
      fstat(artfd, &sbuf);
      bytecnt += sbuf.st_size;

      /* restart the pipeline if batch file is too big... */
      if ((bytecnt > (MAXBATCHSIZ * 2)) && (bytecnt != sbuf.st_size)) {

         /* terminate old pipeline - let child processes finish */
         close(compressio[1]);
         wait((int *) 0);
         wait((int *) 0);

         /* start new pipeline */
         if (do_shelloil(sitename, fp)) {
            close(compressio[0]), close(compressio[1]);
            close(uuxio[0]), close(uuxio[1]);
            fclose(fp);
            printf("Aborting news batching for %s.\n", sitename);
            return;
         }
         bytecnt = sbuf.st_size;
      }

      /* write header line */
      sprintf(buf, "#! rnews %ld\n", (long) sbuf.st_size);
      write(compressio[1], buf, strlen(buf)); 
      bytecnt += strlen(buf);

      /* copy article to compress pipe */
      while (i = read(artfd, buf, sizeof(buf))) {
         if (write(compressio[1], buf, i) != i) {
            printf("Error writing to compress pipe!\n");
            printf("Aborting newsbatching for %s.\n", sitename);
            close(compressio[1]);
            close(artfd);
            fclose(fp);
            return;
         }
      }

      /* close article file */ 
      close(artfd);
   }      

   /* terminate pipeline - let child processes finish */
   close(compressio[1]);
   wait((int *) 0);
   wait((int *) 0);

   /* all done */
   fclose(fp);
   unlink(wrkfile);
   return;
}



do_cmd(sitename, id)
char *sitename; 
char id;
{
   int i, childpid;
   char buf[LINEBUFSIZ];

   if (debugflg)  {
      printf("Execution phase: site = %s, service id = %c\n",
         sitename, id);
   }

   /* chk for built-in services */
   switch (id) {
      case 'p':
         if (!debugflg)
            makepollfile(sitename);
         else 
            printf("   Cmd = makepollfile(%s)\n", sitename);
         return;
      case 'N':
         bit12flg=1;  /* batch news using only 12-bit compress - fall thru! */
         if (!debugflg)
            do_batch(sitename);
         else 
            printf("   Cmd = do_batch(%s)\n", sitename);
         bit12flg=0;  /* batch news using only 12-bit compress - fall thru! */
         return;
      case 'g':
         bitgzip=1;  /* batch news using gzip - fall thru! */
         if (!debugflg)
            do_batch(sitename);
         else 
            printf("   Cmd = do_batch(%s)\n", sitename);
         bitgzip=0;  /* batch news using gzip - fall thru! */
         return;
      case 'h':
         bithgzip=1;  /* batch news using gzip - fall thru! */
         if (!debugflg)
            do_batch(sitename);
         else 
            printf("   Cmd = do_batch(%s)\n", sitename);
         bithgzip=0;  /* batch news using gzip - fall thru! */
         return;
      case 'n':
         if (!debugflg)
            do_batch(sitename);
         else 
            printf("   Cmd = do_batch(%s)\n", sitename);
         return;
      default:
         break;
   }


   /* find service id in services list */
   for (i=0; (i<servcnt) && (id != serv_list[i].id); i++)
      ; 
   if (id != serv_list[i].id) { 
      printf("   No command for service id '%c'!  Aborting.\n", id);
      exit(-1);
   }

   /* Setup and do command.  The command string can have up to
    * four '%s' specs in it, which will be satisfied with sitename...
   */
   sprintf(buf, serv_list[i].cmd, sitename, sitename, sitename, sitename); 
   if (debugflg) 
      printf("   Cmd = %s\n", buf);
   else {
      if (verboseflg)
         printf("   Executing: '%s'\n", buf);
      switch (childpid = fork()) {
         case 0:
            execle("/bin/sh", "sh", "-c", buf, (char *) 0, environ);
            printf("*** ERROR!!! exec failed!\n");
            break;
         case -1:
            printf("*** ERROR: cannot fork! \n");
         default:
            wait((int *) 0);
      }
   }
}



main(argc, argv)
int argc;
char **argv;
{
   extern char *optarg;
   char *p;
   int c, errflg;
   int jobcnt;

   /* process the arguments */
   errflg = 0;
   while ((c = getopt(argc, argv, "ads:v")) != EOF) {
      switch (c) {
         case 'd':
            debugflg++;
            break;
         case 's':
            servlist = optarg;
            break; 
         case 'v':
            verboseflg++;
            break;
         case '?':
            errflg++; 
      }
   }
   if (errflg) {
      print_usage();
      exit(-1);
   }

   /* chatter a bit */
   if (debugflg) 
      printf("Debug flag set.  Service list = %s\n",
         (servlist == (char *) 0) ? "(null)" : servlist);
   if (verboseflg && debugflg) 
      printf("Verbose flag set.\n");

   /* build list of sites with hour entry matching current hour */
   if ((jobcnt = buildlist()) == 0) {
      /* nothing to do */
      if (debugflg || verboseflg)
         printf("No commands to execute.  All done.\n");
      exit(0);
   } 

   /* initialize list of commands for the various services */
   servcnt = getservices();
     
   /* process list */
   while (jobcnt) {

      /* do the command for each service specified */
      for (p=poll_list[jobcnt-1].services; *p; ++p) {

         if (servlist == (char *) 0)
            do_cmd(poll_list[jobcnt-1].sitename, *p);
         else 
            if (strchr(servlist, *p) != NULL)  
               do_cmd(poll_list[jobcnt-1].sitename, *p);
      }
      jobcnt--;
   } 

   if (debugflg || verboseflg)
      printf("All done.\n");
}
